﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Sudoku
{
    /// <summary>
    /// Třída reprezentující sudoku. Poskytuje metody pro posun podle zadání a pro ověření okamžité platnosti políčka.
    /// </summary>
    class Sudoku : ICloneable
    {
        /// <summary>
        /// Tabulka 9x9 do které se vepisují jednostlivá čísla sudoku. První je souřadnice y, druhá x.
        /// </summary>
        public int[][] Tabulka { get; set; }

        /// <summary>
        /// Vytvoří novou instanci sudoku a připraví sloupec tabulky pro zápis řádků.
        /// </summary>
        public Sudoku()
        {
            Tabulka = new int[9][];
        }

        /// <summary>
        /// Vytvoří hlubokou kopii sudoku, implementováno přes rozhraní IClonable
        /// </summary>
        /// <returns>Hlubá kopie daného sudoku</returns>
        public object Clone()
        {
            Sudoku copy = new Sudoku(); //nová instance
            for (int i = 0; i < 9; i++) //postupně se projedou všechny řádky kopírovaného sudoku a zkopírují se do nové instance
            {
                copy.Tabulka[i] = new int[9];
                Tabulka[i].CopyTo(copy.Tabulka[i], 0);
            }

            return copy;
        }

        /// <summary>
        /// Ověří, zdali okamžitá hodnota čísla v políčku koresponduje s ostatními již zapsanými hodnotami
        /// </summary>
        /// <param name="x">souřadnice x (horizontální) ověřovaného políčka</param>
        /// <param name="y">souřadnice y (vertikální) ověřovaného políčka</param>
        /// <returns>Zdali hodnota koresponduje.</returns>
        public bool FieldMatches(int x, int y)
        {
            bool result = true; //navrátová proměnná

            if ((x < 0 && x > 8) || (y < 0 && y > 8)) result = false; //pokud souřadnice nejsou v rozsahu indexu tabulky, nepokračujeme

            if (result) //ověření řádku
            {
                List<int> radek = new List<int>();
                for (int naradku = 0; naradku < 9; naradku++) if(Tabulka[y][naradku] > 0) radek.Add(Tabulka[y][naradku]); //do kolekce čísel z řádku přidáme všechna kromě nul
                
                result = radek.Distinct().Count() == radek.Count(); //Odstraněním duplikátů zjistíme, zdali jsou všechna čísla unikátní (dle pravidel)
            }

            if (result) //ověření sloupce
            {
                List<int> sloupec = new List<int>();
                for (int vesloupci = 0; vesloupci < 9; vesloupci++) if (Tabulka[vesloupci][x] > 0) sloupec.Add(Tabulka[vesloupci][x]); //do kolekce čísel ze sloupce přidáme všechna kromě nul

                result = sloupec.Distinct().Count() == sloupec.Count(); //Odstraněním duplikátů zjistíme, zdali jsou všechna čísla unikátní (dle pravidel)
            }

            if (result) // ověření čtverce 3x3
            {
                List<int> ctverec = new List<int>();
                int nasobicx = (int)(Math.Floor(x / 3.0) * 3); //zjistíme souřadnici x velkého čtverce (0, 1, 2) a vynásobíme 3x (kvůli posunu)
                int nasobicy = (int)(Math.Floor(y / 3.0) * 3); //zjistíme souřadnici x velkého čtverce (0, 1, 2) a vynásobíme 3x (kvůli posunu)
                for (int yctverce = 0; yctverce < 3; yctverce++) for (int xctverce = 0; xctverce < 3; xctverce++) if (Tabulka[yctverce + nasobicy][xctverce + nasobicx] > 0) ctverec.Add(Tabulka[yctverce + nasobicy][xctverce + nasobicx]); //do kolekce čísel ze čtverce přidáme všechna kromě nul

                result = ctverec.Distinct().Count() == ctverec.Count(); //Odstraněním duplikátů zjistíme, zdali jsou všechna čísla unikátní (dle pravidel)
            }

            return result;
        }

        /// <summary>
        /// Posune aktuální souřadnice na ty následující (ve směru doprava a dolů), které byly v zadání prázdné
        /// </summary>
        /// <param name="x">aktuální souřadnice x</param>
        /// <param name="y">aktuální souřadnice y</param>
        /// <param name="outx">nová souřadnice x</param>
        /// <param name="outy">nová souřadnice y</param>
        /// <param name="zadani">Sudoku zadání, z něhož vyčteme zadaná čísla, která nelze měnit</param>
        /// <returns>Zdali již nové souřadnice neleží mimo tabulku, značí úspěšné vyřešení</returns>
        public bool NextField(int x, int y, out int outx, out int outy, Sudoku zadani)
        {
            bool result = true, iszero = false; //jsou souřadnice mimo hranice tabulky? | je vybrané číslo v zadání nula?

            while (result && !iszero) //hledáme, dokud nenarazíme na políčko ze zadání s nulou nabo na konec tabulky
            {
                x++;
                if (x > 8)
                {
                    x = 0;
                    y++;
                }

                if (y > 8) result = false;
                else if (zadani.Tabulka[y][x] == 0) iszero = true;
            }

            outx = x; //přiřadíme výstupové proměnné
            outy = y;
            return result;
        }

        /// <summary>
        /// Posune aktuální souřadnice na předchozí (ve směru doleva a nahoru), které byly v zadání prázdné
        /// </summary>
        /// <param name="x">aktuální souřadnice x</param>
        /// <param name="y">aktuální souřadnice y</param>
        /// <param name="outx">nová souřadnice x</param>
        /// <param name="outy">nová souřadnice y</param>
        /// <param name="zadani">Sudoku zadání, z něhož vyčteme zadaná čísla, která nelze měnit</param>
        /// <returns>Zdali již nové souřadnice neleží mimo tabulku, značí že je sudoku neřešitelné</returns>
        public bool PreviousField(int x, int y, out int outx, out int outy, Sudoku zadani)
        {
            bool result = true, iszero = false; //jsou souřadnice mimo hranice tabulky? | je vybrané číslo v zadání nula?

            while (result && !iszero) //hledáme, dokud nenarazíme na políčko ze zadání s nulou nabo na konec tabulky
            {
                x--;
                if (x < 0)
                {
                    x = 8;
                    y--;
                }

                if (y < 0) result = false;
                else if (zadani.Tabulka[y][x] == 0) iszero = true;
            }

            outx = x; //přiřadíme výstupové proměnné
            outy = y;
            return result;
        }

         /// <summary>
         /// Vypíše sudoku (prázdná místa jako nuly)
         /// </summary>
         /// <returns>Vypsané sudoku jako formátovaný string</returns>
        public override string ToString()
        {
            string table = ""; //návratová proměnná

            foreach (int[] bb in Tabulka) //ve dvou vnořených iteračních cyklech projedeme a připíšeme všechna čísla tabulky, přičemž je oddělíme mezerami a odřádkováními
            {
                foreach (int b in bb) table += b + " ";
                table += "\n";
            }

            return table;
        }
    }
}
